# vim: foldmethod=marker:foldmarker=\ {{{,\ }}}:cms=#%s

import os
import json
import pytest
from requests import codes
from guide_search.esearch import Esearch, Dsl
from mock import Mock
import requests
from utils import log, AttrDict
from guide_search.exceptions import (
    JSONDecodeError,
    BadRequestError,
    ResourceNotFoundError,
    ConflictError,
    PreconditionError,
    ServerError,
    ElasticsearchError,
    ServiceUnreachableError,
    UnknownError,
    ResultParseError,
    CommitError,
    ValidationError, )

@pytest.fixture
def esearch():
    return Esearch(host='http://', control='knowledge_config/control')



def test_unit_dsl():
    """
        Dsl.dsl() returns a well formed query.
    """
    # mocks /guide_ui/server.py::parse_search
    search = {}
    search['search'] = 'mermaid'
    search['sort_type'] = 'score'
    search['page_size'] = 25
    search['page_number'] = 1
    search['fields'] = ["id", "scope", "keywords^5", "title^5"]
    search['order'] = "desc"
    search['filter'] = []

    # mocks /guide_ui/server.py::get_search
    sortmap = {"score": "_score", "popularity": "popularity", "date": "master.lastmodified"}
    sort = sortmap.get(search['sort_type'], "_score")
    dsl = Dsl()
    dsl.ui_query(multi_match=search['search'],
       facets=search['filter'],
       fields=search['fields'])
    dsl.size(search['page_size'], search['page_number'])
    dsl.sort(sort, search['order'])
    dsl.facets()
    dsl.suggest(search['search'], "scope")

    expected = {'aggs': {'nest': {'aggs': {'facetnames': {'aggs': {'focinames': {'terms': {'field': 'facets.foci'}}}, 'terms': {'field': 'facets.facet'}}}, 'nested': {'path': 'facets'}}}, 'query': {'bool': {'filter': [{'bool': {'filter': []}}], 'must': {'multi_match': {'fields': ['id', 'scope', 'keywords^5', 'title^5'], 'query': 'mermaid'}}}}, 'size': 25, 'sort': [{'_score': {'order': 'desc'}}], 'suggest': {'didyoumean': {'term': {'field': 'scope', 'size': 10}, 'text': 'mermaid'}}} 
    assert(dsl.dsl() == expected) and log('dsl returns a well-formed query')


def test_unit_match_all():
    dsl= Dsl().match_all()
    expected = {'query': {'bool': {'must': [{'match_all': {}}]}}}
    assert(dsl.dsl() == expected) and log('match_all returns a well-formed query')


def test_unit_more_like_this():
    index_name='dummy'
    dsl= Dsl().more_like_this(index_name, source= 'franchise_fr')
    expected = {'_source': ['id', 'title', 'scope'], 'query': {'more_like_this': {'fields': ['keywords', 'scope', 'markup'], 'like': [{'_id': 'franchise_fr', '_index': index_name, '_type': 'articles'}], 'max_query_terms': 12, 'min_term_freq': 1}}}
    assert(dsl.dsl() == expected) and log('more_like_this returns a well-formed query')


@pytest.mark.parametrize("facet_filter", ["simple", "complex"])
def test_unit_m_facet_filter(facet_filter):
    root=os.path.dirname(__file__)
    filepath = os.path.join(root, 'fixtures', 'buildFacetFilter', facet_filter + '.input')
    with open(filepath) as f:
        foci_selected = json.load(f)

    filepath = os.path.join(root, 'fixtures', 'buildFacetFilter', facet_filter + '.result')
    with open(filepath) as f:
        expected = json.load(f)
    assert Dsl().m_facet_filter(foci_selected) == expected  and log('facet_filter  %s formed as expected' % facet_filter)


def test_unit_makeurl(esearch):
    url = esearch.make_url('dev', "articles", "bank")
    expected = '{}/dev/articles/bank'.format(esearch.host)
    assert url == expected and log('make_url returns a well-formed URL')


def MockGetRequests( status_code, content):
    """ mock Requests object for esearch """
    Requests = Mock()
    response = AttrDict({ 'status_code' : status_code, 'content' : content, 'text':'sometext' })
    response.json = lambda : json.loads(content)
    Requests.get.return_value=response
    return Requests


@pytest.mark.parametrize("status_code", [codes.OK, codes.CREATED])
def test_unit_request_blank_string(esearch, status_code):
    esearch.Requests = MockGetRequests(status_code = status_code, content  = '')
    assert esearch.request(method='get',url='http://google.com') == True
    log('esearch.request returns True if response is OK/CREATED but content is blank')


@pytest.mark.parametrize("status_code", [codes.OK, codes.CREATED])
def test_unit_request_bad_json(esearch, status_code):
    esearch.Requests = MockGetRequests(status_code = status_code, content  = '{')
    with pytest.raises(JSONDecodeError):
        esearch.request(method='get',url='http://google.com')
    log(' esearch.request raises exception if response is OK/CREATED but content is bad json ')


def test_unit_request_bad_400(esearch):
    esearch.Requests = MockGetRequests(status_code = codes.BAD, content  = '{}')
    with pytest.raises(BadRequestError):
        assert esearch.request(method='get',url='http://google.com')
    log(' esearch.request raises exception if response is BAD REQUEST')


def test_unit_request_bad_404(esearch):
    esearch.Requests = MockGetRequests(status_code = codes.NOT_FOUND, content  = '{}')
    with pytest.raises(ResourceNotFoundError):
        assert esearch.request(method='get',url='http://google.com')
    log(' esearch.request raises exception if response is BAD REQUEST')


def test_unit_request_bad_409(esearch):
    esearch.Requests = MockGetRequests(status_code = codes.CONFLICT, content  = '{}')
    with pytest.raises(ConflictError):
        assert esearch.request(method='get',url='http://google.com')
    log(' esearch.request raises exception if response is CONFLICT')


def test_unit_request_bad_412(esearch):
    esearch.Requests = MockGetRequests(status_code = codes.PRECONDITION_FAILED, content  = '{}')
    with pytest.raises(PreconditionError):
        assert esearch.request(method='get',url='http://google.com')
    log(' esearch.request raises exception if response is PRECONDITION_FAILED')


def test_unit_request_bad_500(esearch):
    esearch.Requests = MockGetRequests(status_code = codes.INTERNAL_SERVER_ERROR, content  = '{}')
    with pytest.raises(ElasticsearchError):
        assert esearch.request(method='get',url='http://google.com')
    log(' esearch.request raises exception if response is INTERNAL_SERVER_ERROR')


def test_unit_request_bad_503(esearch):
    esearch.Requests = MockGetRequests(status_code = codes.SERVICE_UNAVAILABLE, content  = '{}')
    with pytest.raises(ServiceUnreachableError):
        assert esearch.request(method='get',url='http://google.com')
    log(' esearch.request raises exception if response is SERVICE_UNAVAILABLE')





# @pytest.fixture
# def root():
    # return os.path.dirname(__file__)

# @pytest.fixture
# def index():
    # return 'dev'

# @pytest.fixture
# def control():
    # return 'knowledge_config/control'

# @pytest.fixture
# def eserver():
    # return 'gl-know-ap33.lnx.lr.net:9200'

# @pytest.fixture
# def elasticsearch_version():
    # return '1.4.5'


# @pytest.fixture
# def esearch(eserver, control, index):
    # return Esearch(host='http://{}'.format(eserver),control=control)

# @pytest.fixture
# def dsl():
    # return Dsl()



# def test_buildPage(esearch, dsl):
    # """
    # page correctly populates a dictionary of request fields based on:
        # - the required page number (page)
        # - number of results to be displayed per page (size)

    # https://www.elastic.co/guide/en/elasticsearch/reference/5.1/search-request-from-size.html

    # """
    
    # assert esearch.page() == {'size': 5000}
    # assert esearch.page(page=0) == {'size': 5000}
    # assert esearch.page(page=1) == {'size': 5000}
    # assert esearch.page(page=2) == {'size': 5000}
    # assert esearch.page(size=10) == {'size': 10}
    # assert esearch.page(size=10) == {'size': 10}
    # assert esearch.page(size=10, page=0) == {'size': 10}
    # assert esearch.page(size=10, page=1) == {'size': 10}
    # assert esearch.page(size=10, page=2) == {'size': 10, 'from' : 10 }
    # assert esearch.page(size=10, page=20) == {'size': 10, 'from' : 190 }
    # assert esearch.page(size=10, page=-1) == {'size': 10}
    # assert esearch.page(size=-10, page=1) == {'size': -10} # GE : TODO is this desired?

# def test_buildSort(esearch):

    # """
    # buildSort correctly populates a dictionary of request fields based on:
        # - sort = search_type = 'score', 'date', or 'popularity'
        # - order = 'asc' or 'desc'

    # https://www.elastic.co/guide/en/elasticsearch/reference/current/search-request-sort.html

    # """

    # assert dsl.buildSort() == {'sort': [{'_score': {'order': 'desc'}}]}
    # assert dsl.buildSort(order='desc') == {'sort': [{'_score': {'order': 'desc'}}]}
    # assert dsl.buildSort(sort='score') == {'sort': [{'_score': {'order': 'desc'}}]}
    # assert dsl.buildSort(sort='score',order='desc') == {'sort': [{'_score': {'order': 'desc'}}]}
    # assert dsl.buildSort(sort='date',order='asc') == {'sort': [{'master.lastmodified': {'order': 'asc'}}]}

# def test_buildDsl(esearch, index):
    # """
    # buildDsl returns a well formed elasticsearch query

    # """
    # fields = ["id", "title", "scope", "master.document", "master.version", "sensitivity", "purpose"]
    # expected = {
        # 'fields': ['id',
        # 'title',
        # 'scope',
        # 'master.document',
        # 'master.version',
        # 'sensitivity',
        # 'purpose'],
        # 'query': {
            # 'filtered': {'filter': {'bool': {'must': []}},
            # 'query': {'bool': {'should': []}}}
            # },
        # 'size': 5000,
        # 'sort': 'id'}
    # dsl = esearch.buildDsl(fields=fields, index=index)
    # assert dsl == expected
    # # this should now work: esearch.search(index,"articles",dsl)




# def test_get_documents(esearch,index):
    # """
    # get_documents returns a whole load of articles
    # """
    # r = esearch.get_documents(type="article", index=index)
    # assert r['total'] > 1


# # TODO
# # def test_getAssociateList(esearch, index):
    # # r = esearch.getAssociateList(index, "article")
    # # pass


# def test_get_articles(esearch, index):
    # """
    # get_articles returns a whole load of articles
    # """
    # r = esearch.get_articles(index=index)
    # assert len(r) > 1


# def test_get_items(esearch,index):
    # r = esearch.get_items(index=index)
    # assert len(r) > 1


# def test_get_snippets(esearch,index):
    # """
    # """
    # r = esearch.get_snippets(index=index)
    # assert len(r) > 1


# def test_get_document(esearch, index):
    # """
    # """
    # r = esearch.get_document(index=index, type='article', id='fr_gen_reg_lra2002')
    # assert 'First registration' in r['title']


# def test_get_control(esearch):
    # """
    # """
    # r = esearch.get_control(item='styles')
    # assert 'legal_bold' in r.keys()

# def test_get_similar(esearch, index):
    # """
    # """
    # r = esearch.get_similar(index=index, keywords=['first','registration'])
    # ids = [ i['id'][0] for i in r ]
    # expected = ['admin_receiv_adrec', 'boundaries_electro', 'leases_regd_mortgage']
    # assert not set(ids).difference(set(expected))

# @pytest.mark.parametrize('id', ['work_all_tm_ji'])
# def test_get_cluster(esearch, index,id):
    # """
    # """
    # r = esearch.get_cluster(index=index, id=id, size=1)
    # assert id in r[0]['id']

# def test_get_cluster_ist(esearch, index):
    # """
    # """
    # r = esearch.get_cluster_list(index=index)
    # assert len(r) > 100

# # def test_getWhere(esearch, index):
    # # r = esearch.getWhere(index=index, item='styles')
    # # import IPython; IPython.embed()

# # def test_clearIndex ():
    # # raise Exception('TODO')

# # def test_postLog(esearch, index):
    # # seq = '' update["seq"]
    # # url = esearch.make_url(index, "commits", "{0}-{1}".format(index, seq)).strip()
    # # r = esearch.request("get", url)
    # # import IPython; IPython.embed()


# # def test_getSuccess():
    # # raise Exception('TODO')

# # def test_getStatus():
    # # raise Exception('TODO')

# # def test_postEnd():
    # # raise Exception('TODO')

# # def test_resetUp():
    # # raise Exception('TODO')

# # def test_postUp():
    # # raise Exception('TODO')

# # def test_postKM():
    # # raise Exception('TODO')

# # def test_delKM():
    # # raise Exception('TODO')






# Guide specific exceptions
# JP: TODO think about whether exceptions should all be logged


class BaseGuideError(Exception):
    """
    Base exception class for Guide .
    """
    def __init__(self, *args, **kwargs):
        super(BaseGuideError, self).__init__(*args, **kwargs)


class CommitError(BaseGuideError):
    def __init__(self, message):
        super(CommitError, self).__init__(message)


class ValidationError(BaseGuideError):
    def __init__(self, message):
        super(ValidationError, self).__init__(message)


class JSONDecodeError(BaseGuideError):
    """
    Unable to decode received JSON.
    """
    def __init__(self, response):
        self.response = response
        super(JSONDecodeError, self).__init__(
            'Unable to decode received JSON, you can inspect exception\'s '
            '"response" attribute to find out what the response was'
        )

# ElasticSearch error returns
# JP: I couldn't find very good documentation for these TODO maybe need to look harder for docs and revise these
# JP TODO also think about security implications of error returns and whether these get back to the client


class BadRequestError(BaseGuideError):  # 400
    def __init__(self, message):
        super(BadRequestError, self).__init__(message)


class ResourceNotFoundError(BaseGuideError):    # 404
    def __init__(self):
        super(ResourceNotFoundError, self).__init__("Requested resource doesn't exist")


class ConflictError(BaseGuideError):    # 409
    def __init__(self):
        super(ConflictError, self).__init__("Resource version on the server is newer than on the client")


class PreconditionError(BaseGuideError):    # 412
    def __init__(self):
        super(PreconditionError, self).__init__("Precondition Error")


class ServerError(BaseGuideError):  # 500
    def __init__(self):
        super(ServerError, self).__init__("ServerError Error")

class ElasticsearchError(BaseGuideError):  # 500
    def __init__(self, message):
        super(ElasticsearchError, self).__init__("Elasticsearch Error {0}".format(message))



class ServiceUnreachableError(BaseGuideError):  # 503
    def __init__(self):
        super(ServiceUnreachableError, self).__init__("ServerError Error")


class ResultParseError(BaseGuideError):
    def __init__(self, message):
        super(BadRequestError, self).__init__("ElasticSearch return not in expected format {}".format(message))


class UnknownError(BaseGuideError):
    def __init__(self, message):
        super(UnknownError, self).__init__("Unknown error from elastic search{0}".format(message))



import requests
import json
from datetime import datetime, timezone
from time import sleep
from guide_search._version import __version__

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin
from posixpath import join as posixjoin

from guide_search.dsl import Dsl
from guide_search.exceptions import (
    JSONDecodeError,
    BadRequestError,
    ResourceNotFoundError,
    ConflictError,
    PreconditionError,
    ServerError,
    ElasticsearchError,
    ServiceUnreachableError,
    UnknownError,
    ResultParseError,
    CommitError,
    ValidationError,
    )


"""
guide-search

interface to the guidance index in elasticsearch

currently built for elasticseach v1.4


"""

# TODO JP - need to put dict decodes inside try-except and return an appropriate error message if elasticsearch
# responce is not as expected. Currently error message is unhelpful.


class Esearch(object):

    def __init__(self, host,  control, Requests=requests, **kwargs):
        self.init(host, control, **kwargs)
        self.Requests=Requests

    def init(self, host, control, **kwargs):
        self.host = host
        if len(control) == 0:
            self.control = 'guide_config'
        else:
            control = control.split('/')
            for x in control:
                if len(x) > 0:
                    self.control = x
                    break

        self.logger = kwargs.get("logger", None)
        self.timeout = kwargs.get("timeout", 5)
        self.retries = kwargs.get("retries", 3)
        self.reset_session()

    def reset_session(self):
        if self.logger:
            self.logger.debug("reset elasticsearch connection and wait 5 secs")

        try:
            self.session.close()
            del self.session
        except:
            pass
        sleep(5) # give ES some time
        self.session = requests.Session() # create a session to keep a TCP connection open
        if self.logger:
            self.logger.debug("elastic search connection reset")

# In principle we should be using the Elasticsearch maintained python library to do requests
# however this code has worked well for a long time and mitigates the ES hanging problem we saw
    def request(self, method, url, params=None, data=None, timeout=None):
        """Makes requests to elasticsearch and returns result in json format"""
        kwargs = dict({}, **{
            'headers': {},
        })
        if params:
            kwargs['params'] = params

        # note : elastic search takes a payload on
        # GET and delete requests - this is non standard
        if data:
            if isinstance(data, str): # bulk updates are newline deliminated json
                kwargs['data'] = data
                kwargs['headers']['Content-Type'] = 'application/x-ndjson'
            else:
                kwargs['data'] = json.dumps(data)
                kwargs['headers']['Content-Type'] = 'application/json'



        # this retry loop assumes that all requests are idempotent and so we can retry on fail
        retries = 0
        while True:
            try:
                if retries > 0:
                    sleep(0.01)  # reset thread scheduling 0.0001?
                response = getattr(self.session, method)(url,timeout=self.timeout,**kwargs)
                break
            except Exception as e:
                if self.logger:
                    self.logger.debug("esearch retrying {0} {1} {2}".format(retries, method, url))
                retries = retries + 1
                if retries == self.retries:
                    # the underlying libraries should normally automatically re-establish a connection
                    # this is to deal with a situation where ES has hung
                    self.reset_session()
                if retries > self.retries:
                    raise e

        if response.status_code in (200, 201):
            if not response.content.strip():  # success but no response content
                return True
            try:
                return response.json()
            except (ValueError, TypeError):
                raise JSONDecodeError(response)

        elif response.status_code == 400:
            raise BadRequestError(response.text)
        elif response.status_code == 404:
            raise ResourceNotFoundError
        elif response.status_code == 409:
            raise ConflictError
        elif response.status_code == 412:
            raise PreconditionError
        elif response.status_code == 500:
            raise ElasticsearchError(response.text)
        elif response.status_code == 503:
            raise ServiceUnreachableError
        raise UnknownError(response.status_code)

    def make_url(self, index, *args):
        url = urljoin(self.host, posixjoin(index, *args))
        return url


# JP: TODO refactor this to use the smaller functions in the DSL class
# JP: TODO refactor to make the queries between delivery and management more consistent

    def buildDsl(self, **kwargs):
        dsl = Dsl().match_all()

        if "_fields" in kwargs:
            dsl.fields(kwargs["_fields"])
        dsl.size('5000').sort('id')

        if "keywords" in kwargs:
            dsl.Q_should({"terms": {"keywords": kwargs["keywords"],
                                    "minimum_should_match": "-0%"}})  # todo THINK ABOUT DOING MATCH PHRASE HERE
        if "title" in kwargs:
            dsl.Q_should({"match": {"title": {"query": kwargs["title"], "operator": "and"}}})
        if "scope" in kwargs:
            dsl.Q_should({"match": {"scope": {"query": kwargs["scope"], "operator": "and"}}})
        if "markup" in kwargs:
            dsl.Q_should({"match": {"markup": {"query": kwargs["markup"], "operator": "and"}}})
        if "id" in kwargs:
            dsl.Q_should({"match": {"id": {"query": kwargs["id"], "operator": "and"}}})

            # filters
        # JP TODO the following 2 lines were commented out because they caused the Useful Links to fail. However they are probably needed for the editor so need to get to the bottom of this
        if "purpose" in kwargs:
           dsl.Q_filter({"match": {"purpose": kwargs["purpose"]}})

            # nested into clusters # TODO should this be a filter ?
        if "cluster" in kwargs:
            clusters = [{"term": {"clusters.cluster": kwargs["cluster"]}}]
            dsl.Q_filter({"nested": {"path": "clusters", "filter": {"bool": {"must": clusters}}}})

            # nested into facets
        if "facet" in kwargs:
            facets = [{"term": {"facets.facet": kwargs["facet"]}}]
            if "focus" in kwargs:
                facets.append({"term": {"facets.foci": kwargs["focus"]}})
            facetFilter = {"nested": {"path": "facets", "filter": {"bool": {"must": facets}}}}
            dsl.Q_filter(facetFilter)

        return dsl.dsl()


# JP: TODO think about whether to return whole ES response just the hits or something more generic
# query functions
    def search(self, index, type, dsl):
        url = self.make_url(index, type, "_search")
        r = self.request("get", url, data=dsl)
        return r


    def get_documents(self, index, type, **kwargs):
        dsl = self.buildDsl(**kwargs)
        r = self.search(index, type + "s", dsl)
        try:
            results = []
            for hit in r["hits"]["hits"]:
                results.append(hit["_source"])
            return results
        except:
            raise ResultParseError

    def get_associate_list(self, index, aType):
        # JP TODO  this needs to be re-written as it no longe makes sense as associates are items !
        articles = []
        dsl = Dsl().associate_list(aType).dsl()
        res = self.search(index, "articles", dsl)
        for hit in res["hits"]["hits"]:
            articles.append(hit["_source"])
        return {"count": res["hits"]["total"], "articles": articles}

    def get_articles(self, index, **kwargs):
        r = self.get_documents(index, "article", **kwargs)
        return r

    def get_publications(self, index):
        url = self.make_url(index, "record", "_search")
        query = Dsl().match_all().fields(['id']).sort("id", "desc").size("100")
        res = self.request("get", url, data=query.dsl())
        publications = []
        for hit in res["hits"]["hits"]:
            publications.append(hit["_source"]["id"])
        return {"count": res["hits"]["total"], "publications": publications}

    def get_publication(self, index, id):
        url = self.make_url(index, "record",  id)
        res = self.request("get", url)
        try:
            return res["_source"]
        except:
            raise ResultParseError


    def get_items(self, index, **kwargs):
        r = self.get_documents(index, "item",  **kwargs)
        return r

    def get_snippets(self, index, **kwargs):
        r = self.get_documents(index, "snippet", **kwargs)
        return r

    # TODO need to refactor to get_object
    def get_document(self, index, type, id):
        url = self.make_url(index, type + 's', id)
        res = self.request("get", url)
        try:
            return res["_source"]
        except:
            raise ResultParseError

    def get_object(self, index, type, id):
        url = self.make_url(index, type, id)
        res = self.request("get", url)
        try:
            return res["_source"]
        except:
            raise ResultParseError

    def get_control(self, folder, id):
        url = self.make_url(self.control, folder, id)
        res = self.request("get", url)
        try:
            return res["_source"][id]
        except:
            raise ResultParseError

    def get_similar(self, index, id):
        articles = []
        dsl = Dsl().more_like_this(index, id).size(5).dsl()
        url = self.make_url(index, "articles", "_search")
        r = self.request("get", url, data=dsl)
        try:
            for hit in r["hits"]["hits"]:
                articles.append(hit["_source"])
        except:
            ResultParseError
        return articles

    def get_cluster(self, index, id, size=100):
        dsl = Dsl().query_cluster(id).size(100).dsl()
        url = self.make_url(index, 'articles', '_search')
        res = self.request("get", url, data=dsl)
        articles = []

        try:
            for hit in res["hits"]["hits"]:
                articles.append(hit["_source"])
        except:
            ResultParseError
        return articles

    def get_cluster_list(self, index, **kwargs):
        if "max" in kwargs:
            max = kwargs["max"]
        else:
            max = 10000

        dsl = Dsl().cluster_list(max).dsl()
        url = self.make_url(index, "articles", "_search")
        res = self.request("get", url, data=dsl)

        try:
            # note I need a ordered dict that contains a list of clusters and their attributes"
            cls = dict([(c["key"], {"id": c["key"], "count": c['doc_count'], "title": "_no landing page",
                                    "scope": "no landing page for cluster " + c["key"]}) for c in
                        res["aggregations"]["clusters"]["clusters"]["buckets"]])

            cls = dict([(c["key"], {"id": c["key"], "count":c['doc_count'], "title":"_no landing page",
                                    "scope":"no landing page for cluster " + c["key"],
                                    "sens":c["sens"]["sens"]["buckets"]}) for c in
                        res["aggregations"]["clusters"]["clusters"]["buckets"]])
        except:
            ResultParseError

        dsl = Dsl().landing_list().size(max).dsl()
        url = self.make_url(index, "articles", "_search")

        res = self.request("get", url, data=dsl)

        try:
            cl = res["hits"]["hits"]
            for c in cl:
                cls[c["_source"]["id"]]["title"] = c["_source"]["title"]
                cls[c["_source"]["id"]]["scope"] = c["_source"]["scope"]
        except:
            ResultParseError

        return cls

    # editing stuff
    #
    def get_item_use(self, index, id):
        dsl = Dsl().item_use(id).dsl()
        url = self.make_url(index, 'articles', '_search')
        deps = []
        res = self.request("get", url, data=dsl)
        try:
            for hit in res["hits"]["hits"]:
                deps.append(hit["_source"]["id"])
        except:
            ResultParseError
        return deps

    def get_snippet_use(self, index, snippet_type, snippet_id):
        url = self.make_url(index, "articles", "_search")
        dsl = (Dsl().
               fields(['id']).
               sort('_score', 'desc').
               Q().
               Q_must({"term": {"snips.snip_type": snippet_type}}).
               Q_must({"term": {"snips.snip_id": snippet_id}}).
               dsl())
        res = self.request("get", url, data=dsl)
        articles = []
        for hit in res["hits"]["hits"]:
            articles.append(hit["_source"]["id"])
        return {"count": res["hits"]["total"], "articles": articles}



    # publishing stuff
    #
    # clear index and set commit record to empty
    def clear_index(self, index):
        query = Dsl().match_all().sort("seq", "desc").size("1")
        url = self.make_url(index, "commits", "_search")
        res = self.request("post", url, data=query.dsl())["hits"]["hits"]
        if len(res) == 1:
            last = res[0]["_source"]
            seq = int(last["seq"]) + 1
        else:
            seq = 0

        self.get_empty(index, seq)  # create a reference to empty then delete all content
        self.del_object(index, "articles", "*")  # wildcard to delete all items but not the document mapping
        self.del_object(index, "items", "*")
        self.del_object(index, "snippets", "*")

    def post_log(self, index, seq, log):
        try:
            url = self.make_url(index, "commits", "{0}-{1}".format(index, seq)).strip()

            r = self.request("get", url)

            last = r["_source"]

            if last['state'] not in ['updating']:
                return
            up = {}
            up["log"] = last["log"]
            up["log"].append(log)
            url = self.make_url(url, "_update")
            self.request("post", url, data={"doc": up, "doc_as_upsert": "true"}, params={"refresh": "true"})
        except:
            pass

    def post_progress(self, index, seq, progress):
        try:
            url = self.make_url(index, "commits", "{0}-{1}".format(index, seq)).strip()

            r = self.request("get", url)

            last = r["_source"]

            if last['state'] not in ['updating']:
                return
            up = {}
            up['log'] = json.dumps(progress) # JP TODO would be nicer if this was a object but this would require a mapping change
            url = self.make_url(url, "_update")
            self.request("post", url, data={"doc": up, "doc_as_upsert": "true"}, params={"refresh": "true"})
        except:
            pass


    # TODO think about combining this with the clear function
    def get_empty(self, index, seq=0):
        up = {}
        # create a commit pointing to empty
        up["seq"] = seq
        up["start"] = datetime.now(timezone.utc).astimezone().isoformat()
        up["finish"] = up["start"]
        up["oldcommit"] = "empty"
        up["newcommit"] = "empty"
        up["state"] = "empty"
        up["log"] = ["no commit yet"]
        url = self.make_url(index, "commits", "{0}-{1}".format(index, up["seq"]))
        self.request("post", url, data=up)
        return up

    def get_last_success(self, index):
        # query = {"query": {
        #             "filtered": {
        #                "query": {
        #                 "match_all": {}},
        #                "filter": {"term": {"state": "complete"}}
        #                 }
        #             },
        #          "sort": {"seq": "desc"},
        #          "size": "1"}  # es1
        query = Dsl().match_all({"term": {"state": "complete"}}).sort("seq", "desc").size("1")
        url = url = self.make_url(index, "commits", "_search")
        # r = self.request("post", url, data=query)["hits"]["hits"] # es 1
        res = self.request("post", url, data=query.dsl())
        states = res["hits"]["hits"]
        if len(states) == 0:
            return self.get_empty(index)
        else:
            return states[0]["_source"]

    def get_state(self, index, seq=None):
        if seq:
            url = self.make_url(index, "commits", "{0}-{1}".format(index, seq))
            res = self.request("get", url,)
            return res["_source"]
        else:
            query = Dsl().match_all().sort("seq", "desc").size(1)
            url = self.make_url(index, "commits", "_search")
            res = self.request("post", url, data=query.dsl())
            states = res["hits"]["hits"]
            if len(states) == 0:
                return self.get_empty(index)
            else:
                return states[0]["_source"]

    def post_end(self, index, seq, newcommit, state, log=None):
        last = self.get_state(index)
        if not seq == last["seq"]:
            m = "Requested seq <{0}> does not match in progress seq <{1}>".format(seq, last["seq"])

            raise CommitError(m)
        elif not last["state"] == "updating":
            m = "Latest commit <{0}> in wrong state = {1}".format(last["seq"], last['state'])

            raise CommitError(m)
        elif not last["newcommit"] == newcommit:
            m = "Requested commit <{0}> does not match in progress commit <{1}>".format(newcommit, last["newcommit"])

            raise CommitError(m)
        else:
            last["finish"] = datetime.now(timezone.utc).astimezone().isoformat()
            last["state"] = state
            # if log:
            #     last["log"] = [log]
            # else:
            #     last["log"] = ""
            url = self.make_url(index, "commits", "{0}-{1}".format(index, seq))
            self.request("post", url, data=last, params={"refresh": "true"})
        return last

    def reset_up(self, index, seq):
        url = self.make_url(index, "commits", "{0}-{1}".format(index, seq))
        self.request("delete", url)

    def post_up(self, index, last, newcommit):
        up = {}
        # start new commit
        seq = int(last["seq"]) + 1
        up["seq"] = seq
        up["start"] = datetime.now(timezone.utc).astimezone().isoformat()
        up["finish"] = datetime.isoformat(datetime.min) + 'Z'
        up["oldcommit"] = last["newcommit"]
        up["newcommit"] = newcommit
        up["state"] = "updating"
        up["log"] = ["updating from {0} to {1} ".format(up["oldcommit"], newcommit)]
        url = self.make_url(index, "commits", "{0}-{1}".format(index, seq), "_update")
        self.request("post", url, data={"doc": up, "doc_as_upsert": "true"}, params={"refresh": "true"})
        return up


    # POST to partially update or completetly create - not idempotent - not safe FIXME TODO
    def post_object(self, index, oType, kmj, upsert=False):
        try:
            id_value = kmj["id"]
        except:
            raise ValidationError("no id field in json structure")

        if upsert:
            # create or update changed fields
            url = self.make_url(index, oType, id_value, "_update")
            # ES5 : True was unquoted changed this to "true"  - ES1 break ?
            doc = {"doc": kmj, "doc_as_upsert": "true"}
        else:
            # completely replace
            url = self.make_url(index, oType, id_value)
            doc = kmj

        self.request("post", url, data=doc)
        return ({"act": "upsert" if upsert else "replace", "status": "200", "reason": "OK"})

    def del_object(self, index, oType, id=""):
        try:
            # es 1
            # if (id == '*'):
            #     data = {"query": {
            #         "match_all": {}}}  # note this is deprecated and cannot be done in v 2.0 onwards
            #     url = self.make_url(index, oType, "_query")
            # else:
            #     data = {}
            #     url = self.make_url(index, oType, id)

            if (id == '*'):
                delete_query = Dsl().match_all()
                url = self.make_url(index, oType, "_delete_by_query")
                res = self.request("post", url, data=delete_query.dsl())
                # TODO read the documentation and handle error conditions.
            else:
                url = self.make_url(index, oType, id)
                res = self.request("delete", url)

        except ResourceNotFoundError:
            res = self.Requests.Response()
            res.status_code = 200
            res.reason = "Resource to be deleted did not exist <{0}>".format(url)
        return res

    def post_bulk(self, channel, bulk):
        url = self.make_url(channel, "_bulk")
        res = self.request("put", url, data = bulk, timeout=3600)
        return res

# setup stuff
    def put_mapping(self, index, mapping, force=False):
        url = self.make_url(index)
        if force:
            try:
                self.request("delete", url)
            except:
                pass
        return self.request("put", url, data=mapping)

    def put_control(self, folder, id, control):
        url = self.make_url(self.control, folder, id)
        if id in control:
            return self.request("put", url, data=control)
        else:
            raise ValidationError("url:{0}".format(url))

    def get_info(self):
        info = {"host": self.host,
                "controls": self.control,
                "guide_search_version": __version__
                }
        try:
            url = self.host
            res = self.request("get", url)
            info["version"] = res["version"]["number"]
            info["name"] = res["name"]
            url = self.make_url("_cluster", "health")
            try:
                health = self.request("get", url)
                info["cluster_name"] = health["cluster_name"]
                info["cluster_health"] = health["status"]
                if health["status"] == "green":
                    info["status"] = "alive"
                else:
                    info["status"] = "limping"
            except:
                info["status"] = "not quite alive"
        except:
            info["status"] = "dead"
        return info

    def get_index_health(self, index):
        current = {}
        try:
            publish = self.get_state(index)
            if publish["state"] not in ["complete", "empty"]:
                last = self.get_last_success[index]
                if last["newcommit"] != publish["newcommit"]:
                    publish["last_successful_commit"] = last["newcommit"]
                    publish["last_successful_publish"] = last["finish"]
            del publish["log"]
            current["publish"] = publish
            for doctype in ["articles", "items", "snippets"]:
                url = self.make_url(index, doctype, "_count")
                try:
                    current[doctype] = {'state': 'OK', 'count': self.request("get", url)['count']}
                except:
                    current[doctype] = {'state': 'error'}

        except:
            current["publish"] = {'state': 'error', 'message': 'no publication record'}
        return current



#!/usr/bin/env python
# coding: utf-8
from __future__ import print_function

try:
    # python setup.py test
    import multiprocessing
except ImportError:
    pass


from setuptools import setup, Extension
from distutils.util import convert_path

# vers = {}
cmdclass = {}
about = {}
with open('guide_search/__about__.py') as f:
    exec(f.read(), about)

# ver_path = convert_path('guide_search/_version.py')
# with open(ver_path) as ver_file:
#     exec(ver_file.read(), vers)



try:
    from Cython.Distutils import build_ext
except ImportError:
    have_cython = False
else:
    have_cython = True
    import sys

    # Failsafe compilation originally from simplejson. Author: Bob Ippolito
    # Copied here from markupsafe
    from distutils.errors import CCompilerError, DistutilsExecError
    from distutils.errors import DistutilsPlatformError

    is_jython = 'java' in sys.platform
    is_pypy = hasattr(sys, 'pypy_version_info')

    ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)
    if sys.platform == 'win32' and sys.version_info > (2, 6):
        # 2.6's distutils.msvc9compiler can raise an IOError when failing to
        # find the compiler
        ext_errors += (IOError,)

    class BuildFailed(Exception):
        pass

    class ve_build_ext(build_ext):
        """This class allows C extension building to fail."""
        def run(self):
            try:
                build_ext.run(self)
            except DistutilsPlatformError:
                raise BuildFailed()

        def build_extension(self, ext):
            try:
                build_ext.build_extension(self, ext)
            except ext_errors:
                raise BuildFailed()
            except ValueError as e:
                # this can happen on Windows 64 bit, see Python issue 7511
                if "'path'" in str(e):
                    raise BuildFailed()
                raise

    cmdclass['build_ext'] = ve_build_ext

# patch bdist_wheel
try:
    from wheel.bdist_wheel import bdist_wheel

    REPLACE = (
        'macosx_10_6_intel.'
        'macosx_10_9_intel.'
        'macosx_10_9_x86_64.'
        'macosx_10_10_intel.'
        'macosx_10_10_x86_64'
    )

    class _bdist_wheel(bdist_wheel):
        def get_tag(self):
            tag = bdist_wheel.get_tag(self)
            if tag[2] == 'macosx_10_6_intel':
                tag = (tag[0], tag[1], REPLACE)
            return tag

    cmdclass['bdist_wheel'] = _bdist_wheel
except ImportError:
    pass


def fread(filepath):
    with open(filepath, 'r') as f:
        return f.read()


def run_setup(with_binary):
    ext_modules = []
    setup(
        name=about['__title__'],
        version=about['__version__'],
        url=about['__url__'],
        author=about['__author__'],
        author_email=about['__author_email__'],
        description=about['__description__'],
        long_description=fread('README.rst'),
        license=about['__license__'],
        packages=['guide_search'],
        py_modules=['guide_search.esearch', 'guide_search.dsl', 'guide_search.exceptions', 'guide_search._version'],
        install_requires=['requests'],
        cmdclass=cmdclass,
        ext_modules=ext_modules,
        zip_safe=False,
        platforms='any',
        tests_require=['nose'],
        test_suite='nose.collector',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Web Environment',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: Implementation :: PyPy',
            'Topic :: Software Development :: Libraries :: Python Modules'
        ]
    )


def try_building_extension():
    try:
        run_setup(True)
    except BuildFailed:
        LINE = '=' * 74
        BUILD_EXT_WARNING = 'WARNING: The C extension could not be ' \
                            'compiled, speedups are not enabled.'
        print(LINE)
        print(BUILD_EXT_WARNING)
        print('Failure information, if any, is above.')
        print('Retrying the build without the C extension now.')
        print()
        run_setup(False)
        print(LINE)
        print(BUILD_EXT_WARNING)
        print('Plain-Python installation succeeded.')
        print(LINE)


if have_cython and not(is_jython or is_pypy):
    try_building_extension()
else:
    run_setup(False)
